import contextlib
import time
from enum import Enum
from typing import Union

from rpcclient.darwin.consts import IOHIDDigitizerEventMask, IOHIDDigitizerTransducerType, IOHIDEventField, \
    IOHIDEventFieldDigitizer, kCFAllocatorDefault, kHIDPage_Consumer, kHIDUsage_Csmr_ACSearch, \
    kHIDUsage_Csmr_FastForward, kHIDUsage_Csmr_Menu, kHIDUsage_Csmr_Mute, kHIDUsage_Csmr_Pause, kHIDUsage_Csmr_Play, \
    kHIDUsage_Csmr_PlayOrPause, kHIDUsage_Csmr_Power, kHIDUsage_Csmr_RandomPlay, kHIDUsage_Csmr_Repeat, \
    kHIDUsage_Csmr_Rewind, kHIDUsage_Csmr_VolumeDecrement, kHIDUsage_Csmr_VolumeIncrement
from rpcclient.exceptions import BadReturnValueError


class TouchEventType(Enum):
    TOUCH_DOWN = 0
    TOUCH_UP = 1
    TOUCH_MOVE = 2


TOUCH_EVENT_SLEEP = 0.1


EVENT_TYPE_PARAMS = {
    TouchEventType.TOUCH_DOWN: {
        'event_flags': (IOHIDDigitizerEventMask.kIOHIDDigitizerEventAttribute.value |
                        IOHIDDigitizerEventMask.kIOHIDDigitizerEventTouch.value |
                        IOHIDDigitizerEventMask.kIOHIDDigitizerEventIdentity.value),
        'button_mask': 1,
        'touch': True,
    },
    TouchEventType.TOUCH_MOVE: {
        'event_flags': (IOHIDDigitizerEventMask.kIOHIDDigitizerEventPosition.value |
                        IOHIDDigitizerEventMask.kIOHIDDigitizerEventAttribute.value),
        'button_mask': 1,
        'touch': True,
    },
    TouchEventType.TOUCH_UP: {
        'event_flags': (IOHIDDigitizerEventMask.kIOHIDDigitizerEventAttribute.value |
                        IOHIDDigitizerEventMask.kIOHIDDigitizerEventTouch.value |
                        IOHIDDigitizerEventMask.kIOHIDDigitizerEventIdentity.value),
        'button_mask': 0,
        'touch': False,
    },
}


class Hid:
    """ Control HID devices and simulate events """

    def __init__(self, client):
        self._client = client

    def send_double_home_button_press(self):
        self.send_home_button_press()
        self.send_home_button_press()

    def send_rewind_button_press(self):
        self.send_key_press(kHIDPage_Consumer, kHIDUsage_Csmr_Rewind)

    def send_random_play_button_press(self):
        self.send_key_press(kHIDPage_Consumer, kHIDUsage_Csmr_RandomPlay)

    def send_repeat_button_press(self):
        self.send_key_press(kHIDPage_Consumer, kHIDUsage_Csmr_Repeat)

    def send_fast_forward_button_press(self):
        self.send_key_press(kHIDPage_Consumer, kHIDUsage_Csmr_FastForward)

    def send_play_button_press(self):
        self.send_key_press(kHIDPage_Consumer, kHIDUsage_Csmr_Play)

    def send_pause_button_press(self):
        self.send_key_press(kHIDPage_Consumer, kHIDUsage_Csmr_Pause)

    def send_play_pause_button_press(self):
        self.send_key_press(kHIDPage_Consumer, kHIDUsage_Csmr_PlayOrPause)

    def send_search_button_press(self):
        self.send_key_press(kHIDPage_Consumer, kHIDUsage_Csmr_ACSearch)

    def send_mute_button_press(self):
        self.send_key_press(kHIDPage_Consumer, kHIDUsage_Csmr_Mute)

    def send_home_button_press(self):
        self.send_key_press(kHIDPage_Consumer, kHIDUsage_Csmr_Menu)

    def send_power_button_press(self, interval: int = 0):
        self.send_key_press(kHIDPage_Consumer, kHIDUsage_Csmr_Power, interval=interval)

    def send_volume_down_button_press(self, interval: int = 0):
        self.send_key_press(kHIDPage_Consumer, kHIDUsage_Csmr_VolumeDecrement, interval=interval)

    def send_volume_up_button_press(self, interval: int = 0):
        self.send_key_press(kHIDPage_Consumer, kHIDUsage_Csmr_VolumeIncrement, interval=interval)

    def send_key_press(self, page: int, key_code: int, interval: Union[float, int] = 0):
        self.send_keyboard_event(page, key_code, True)
        if interval:
            time.sleep(interval)
        self.send_keyboard_event(page, key_code, False)

    def send_keyboard_event(self, page: int, key_code: int, down: bool):
        event = self._client.symbols.IOHIDEventCreateKeyboardEvent(kCFAllocatorDefault,
                                                                   self._client.symbols.mach_absolute_time(),
                                                                   page, key_code, down, 0)
        if not event:
            raise BadReturnValueError('IOHIDEventCreateKeyboardEvent() failed')

        self.dispatch(event)

    def send_swipe_right(self):
        self.send_swipe(0.5, 0.5, 1.0, 0.5)

    def send_swipe_left(self):
        self.send_swipe(0.5, 0.5, 0.0, 0.5)

    def send_swipe_up(self):
        self.send_swipe(0.5, 1.5, 0.5, 0.5)

    def send_swipe_down(self):
        self.send_swipe(0.5, 0.5, 0.5, 1.5)

    def send_swipe(self, from_x: float, from_y: float, to_x: float, to_y: float):
        self.send_touch_event(TouchEventType.TOUCH_DOWN, from_x, from_y)
        time.sleep(TOUCH_EVENT_SLEEP)
        self.send_touch_event(TouchEventType.TOUCH_MOVE, to_x, to_y)
        self.send_touch_event(TouchEventType.TOUCH_UP, to_x, to_y)

    def send_touch_event(self, event_type: TouchEventType, x: float, y: float, pressure: float = 0.4):
        params = EVENT_TYPE_PARAMS[event_type]

        timestamp = self._client.symbols.mach_absolute_time()

        event_flags = params['event_flags']
        touch = params['touch']
        button_mask = params['button_mask']

        parent = self.create_digitizer_event(IOHIDDigitizerTransducerType.kIOHIDDigitizerTransducerTypeHand,
                                             0, 0, event_flags, 0, x, y, 0.0, 0.0, 0.0, touch, touch, 0,
                                             timestamp=timestamp)

        child = self.create_digitizer_finger_event(2, 2, event_flags, button_mask, x, y, 0.0, pressure, 0.0,
                                                   touch, touch, 0, timestamp=timestamp)
        self._client.symbols.IOHIDEventAppendEvent(parent, child)
        self.dispatch(parent)

    def create_digitizer_event(self, type_: Union[IOHIDDigitizerTransducerType, int], index: int, identity: int,
                               event_mask: int, button_mask: int, x: float, y: float, z: float, tip_pressure: float,
                               barrel_pressure: float, range_: bool, touch: bool, options: int, timestamp=None):
        if timestamp is None:
            timestamp = self._client.symbols.mach_absolute_time()

        event = self._client.symbols.IOHIDEventCreateDigitizerEvent(
            kCFAllocatorDefault, timestamp, type_, index, identity, event_mask,
            button_mask, x, y, z, tip_pressure, barrel_pressure, range_, touch, options)

        if not event:
            raise BadReturnValueError('IOHIDEventCreateDigitizerEvent() failed')

        self._client.symbols.IOHIDEventSetIntegerValue(event, IOHIDEventField.kIOHIDEventFieldIsBuiltIn, 0)
        self._client.symbols.IOHIDEventSetIntegerValue(
            event, IOHIDEventFieldDigitizer.kIOHIDEventFieldDigitizerIsDisplayIntegrated, 1)
        self._client.symbols.IOHIDEventSetSenderID(event, 0x8000000817319375)

        return event

    def create_digitizer_finger_event(self, index: int, identity: int, event_mask: int,
                                      button_mask: int, x: float, y: float, z: float, tip_pressure: float,
                                      twist: float, range_: bool, touch: bool, options: int, timestamp=None):
        if timestamp is None:
            timestamp = self._client.symbols.mach_absolute_time()

        event = self._client.symbols.IOHIDEventCreateDigitizerFingerEvent(
            kCFAllocatorDefault, timestamp, index, identity, event_mask,
            button_mask, x, y, z, tip_pressure, twist, range_, touch, options)

        if not event:
            raise BadReturnValueError('IOHIDEventCreateDigitizerFingerEvent() failed')

        self._client.symbols.IOHIDEventSetFloatValue(
            event, IOHIDEventFieldDigitizer.kIOHIDEventFieldDigitizerMajorRadius, 0.5)
        return event

    @contextlib.contextmanager
    def create_hid_client(self):
        client = self._client.symbols.IOHIDEventSystemClientCreate(0)

        if not client:
            raise BadReturnValueError('IOHIDEventSystemClientCreate() failed')

        try:
            yield client
        finally:
            self._client.symbols.CFRelease(client)

    def dispatch(self, event):
        with self.create_hid_client() as hid_client:
            self._client.symbols.IOHIDEventSystemClientDispatchEvent(hid_client, event)



from datetime import datetime
from functools import lru_cache
from typing import List, Mapping
from uuid import UUID

from rpcclient.darwin.common import CfSerializable
from rpcclient.darwin.consts import XPC_ARRAY_APPEND
from rpcclient.darwin.symbol import DarwinSymbol
from rpcclient.exceptions import MissingLibraryError, RpcXpcSerializationError
from rpcclient.structs.consts import RTLD_NOW


class XPCObject(DarwinSymbol):
    @property
    def type(self) -> int:
        return self._client.symbols.xpc_get_type(self)


class XPCArray(XPCObject):
    def set_data(self, buf: bytes, index: int = XPC_ARRAY_APPEND) -> None:
        """
        See https://developer.apple.com/documentation/xpc/1505937-xpc_array_set_data?language=objc
        """
        self._client.symbols.xpc_array_set_data(self, index, buf, len(buf))


class XPCDictionary(XPCObject):
    def set_string(self, key: str, value: str) -> None:
        self._client.symbols.xpc_dictionary_set_string(self, key, value)

    def set_int64(self, key: str, value: int) -> None:
        self._client.symbols.xpc_dictionary_set_int64(self, key, value)

    def set_uint64(self, key: str, value: int) -> None:
        self._client.symbols.xpc_dictionary_set_uint64(self, key, value)

    def set_bool(self, key: str, value: bool) -> None:
        self._client.symbols.xpc_dictionary_set_bool(self, key, value)

    def set_data(self, key: str, value: bytes) -> None:
        self._client.symbols.xpc_dictionary_set_data(self, key, value, len(value))

    def set_fd(self, key: str, value: int) -> None:
        self._client.symbols.xpc_dictionary_set_fd(self, key, value)

    def set_uuid(self, key: str, value: UUID) -> None:
        self._client.symbols.xpc_dictionary_set_uuid(self, key, value.bytes)

    def set_dictionary(self, key: str, value: int) -> None:
        self._client.symbols.xpc_dictionary_set_dictionary(self, key, value)

    def set_object(self, obj: XPCObject) -> None:
        self._client.symbols.xpc_dictionary_set_object(self, obj)

    def set_value(self, key: str, obj: XPCObject) -> None:
        self._client.symbols.xpc_dictionary_set_value(self, key, obj)

    def get_string(self, key: str) -> str:
        return self._client.symbols.xpc_dictionary_get_string(self, key).peek_str()

    def get_int64(self, key: str) -> int:
        return self._client.symbols.xpc_dictionary_get_int64(self, key).c_int64

    def get_uint64(self, key: str) -> int:
        return self._client.symbols.xpc_dictionary_get_uint64(self, key).c_uint64

    def get_bool(self, key: str) -> None:
        return self._client.symbols.xpc_dictionary_get_bool(self, key).c_bool

    def get_data(self, key: str) -> bytes:
        with self._client.safe_malloc(8) as p_length:
            return self._client.symbols.xpc_dictionary_get_data(self, key, p_length).peek(p_length[0])

    def get_fd(self, key: str) -> int:
        return self._client.symbols.xpc_dictionary_get_fd(self, key)

    def get_uuid(self, key: str) -> str:
        return self._client.symbols.xpc_dictionary_get_uuid(self, key).peek_str()

    def get_dictionary(self, key: str) -> 'XPCDictionary':
        return XPCDictionary.create(self._client.symbols.xpc_dictionary_get_dictionary(self, key), self._client)

    def get_object(self, key: str) -> XPCObject:
        return XPCObject.create(self._client.symbols.xpc_dictionary_get_object(self, key), self._client)

    def get_value(self, key: str) -> XPCObject:
        return XPCObject.create(self._client.symbols.xpc_dictionary_get_value(self, key), self._client)


class Xpc:
    def __init__(self, client):
        """
        :param rpcclient.darwin.client.DarwinClient client:
        """
        self._client = client
        self._load_duet_activity_scheduler_manager()
        self.sharedScheduler = self._client.symbols.objc_getClass('_DASScheduler').objc_call('sharedScheduler')

    def create_xpc_dictionary(self) -> XPCDictionary:
        return XPCDictionary.create(self._client.symbols.xpc_dictionary_create(0, 0, 0), self._client)

    def create_xpc_array(self) -> XPCArray:
        return XPCArray.create(self._client.symbols.xpc_array_create_empty(), self._client)

    def send_xpc_dictionary(self, service_name: str, message: XPCDictionary) -> XPCDictionary:
        """
        Send a native XPC dictionary to a XPC service synchronously and return result.

        :param service_name: mach service name
        :param message: xpc message to send
        :return: received response
        """
        return XPCDictionary.create(self.send_message_raw(service_name, message), self._client)

    def send_message_using_cf_serialization(
            self, service_name: str, message: CfSerializable, decode_cf: bool = True) -> CfSerializable:
        """
        Send a CFObject serialized over an XPC object to a XPC service synchronously and return reply.

        :param service_name: mach service name
        :param message: xpc message to send
        :param decode_cf: should response be decoded as CFObject-over-XPCObject or a native XPC object
        :return: received response
        """
        message_raw = self.encode_xpc_message_using_cf_serialization(message)
        if message_raw == 0:
            raise RpcXpcSerializationError()
        response = self.send_message_raw(service_name, message_raw)
        if response == 0:
            raise RpcXpcSerializationError()
        return self.decode_xpc_message_using_cf_serialization(
            response) if decode_cf else self.decode_xpc_object_using_cf_serialization(response)

    def send_object_using_cf_serialization(self, service_name: str, message: CfSerializable) -> CfSerializable:
        """
        Send a native XPC object serialized over a CFObject to a XPC service synchronously and return reply.

        :param service_name: mach service name
        :param message: xpc message to send
        :return: received response
        """
        message_raw = self.encode_xpc_object_using_cf_serialization(message)
        if message_raw == 0:
            raise RpcXpcSerializationError()
        response = self.send_message_raw(service_name, message_raw)
        if response == 0:
            raise RpcXpcSerializationError()
        return self.decode_xpc_object_using_cf_serialization(response)

    def decode_xpc_object_using_cf_serialization(self, address: DarwinSymbol) -> CfSerializable:
        """
        Convert XPC object to python object using CF serialization.
        """
        return self._client.symbols._CFXPCCreateCFObjectFromXPCObject(address).py()

    def encode_xpc_object_using_cf_serialization(self, obj: CfSerializable) -> DarwinSymbol:
        """
        Convert python object to XPC object using CF conversion.
        """
        return self._client.symbols._CFXPCCreateXPCObjectFromCFObject(self._client.cf(obj))

    def decode_xpc_message_using_cf_serialization(self, address: DarwinSymbol) -> CfSerializable:
        """
        Convert a CFObject serialized over an XPCObject to python object
        """
        return self._client.symbols._CFXPCCreateCFObjectFromXPCMessage(address).py()

    def encode_xpc_message_using_cf_serialization(self, obj: CfSerializable) -> DarwinSymbol:
        """
        Convert python object to a CFObject serialized over an XPCObject
        """
        return self._client.symbols._CFXPCCreateXPCMessageWithCFObject(self._client.cf(obj))

    def send_message_raw(self, service_name: str, message_raw: DarwinSymbol) -> DarwinSymbol:
        """ Send a RAW xpc object to given service_name and wait reply. """
        conn = self._connect_to_mach_service(service_name)
        return self._client.symbols.xpc_connection_send_message_with_reply_sync(conn, message_raw)

    def force_run_activities(self, activities: List[str]) -> None:
        self.sharedScheduler.objc_call('forceRunActivities:', self._client.cf(activities))

    @property
    def loaded_activities(self) -> Mapping:
        return self._client.preferences.cf.get_dict('com.apple.xpc.activity2', 'root')

    def set_activity_base_date(self, name: str, date: datetime) -> None:
        activity_base_dates = self.loaded_activities['ActivityBaseDates']
        activity_base_dates[name] = date
        self._client.preferences.cf.set('ActivityBaseDates', activity_base_dates, 'com.apple.xpc.activity2', 'root')

    @lru_cache(maxsize=None)
    def _connect_to_mach_service(self, service_name) -> DarwinSymbol:
        conn = self._client.symbols.xpc_connection_create_mach_service(service_name, 0, 0)
        assert conn != 0, 'failed to create xpc connection'
        self._client.symbols.xpc_connection_set_event_handler(conn, self._client.get_dummy_block())
        self._client.symbols.xpc_connection_resume(conn)
        return conn

    def _load_duet_activity_scheduler_manager(self) -> None:
        options = [
            '/System/Library/PrivateFrameworks/DuetActivityScheduler.framework/DuetActivityScheduler',
        ]
        for option in options:
            if self._client.dlopen(option, RTLD_NOW):
                return
        raise MissingLibraryError('failed to load DuetActivityScheduler')



from collections import namedtuple
from dataclasses import dataclass, field
from typing import Mapping

from objc_types_decoder.decode import decode as decode_type
from objc_types_decoder.decode import decode_with_tail

from rpcclient.client import Client

Property = namedtuple('Property', 'name attributes')
PropertyAttributes = namedtuple('PropertyAttributes', 'synthesize type_ list')


def convert_encoded_property_attributes(encoded):
    conversions = {
        'R': lambda x: 'readonly',
        'C': lambda x: 'copy',
        '&': lambda x: 'strong',
        'N': lambda x: 'nonatomic',
        'G': lambda x: 'getter=' + x[1:],
        'S': lambda x: 'setter=' + x[1:],
        'd': lambda x: 'dynamic',
        'W': lambda x: 'weak',
        'P': lambda x: '<garbage-collected>',
        't': lambda x: 'encoding=' + x[1:],
    }

    type_, tail = decode_with_tail(encoded[1:])
    attributes = []
    synthesize = None
    for attr in filter(None, tail.lstrip(',').split(',')):
        if attr[0] in conversions:
            attributes.append(conversions[attr[0]](attr))
        elif attr[0] == 'V':
            synthesize = attr[1:]

    return PropertyAttributes(type_=type_, synthesize=synthesize, list=attributes)


@dataclass
class Method:
    name: str
    client: Client = field(compare=False)
    address: int = field(compare=False)
    imp: int = field(compare=False)
    type_: str = field(compare=False)
    return_type: str = field(compare=False)
    is_class: bool = field(compare=False)
    args_types: list = field(compare=False)

    @staticmethod
    def from_data(data: Mapping, client):
        """
        Create Method object from raw data.
        :param data: Data as loaded from get_objectivec_symbol_data.m.
        :param rpcclient.darwin.client.DarwinClient client: Darwin client.
        """
        return Method(
            name=data['name'],
            client=client,
            address=client.symbol(data['address']),
            imp=client.symbol(data['imp']),
            type_=data['type'],
            return_type=decode_type(data['return_type']),
            is_class=data['is_class'],
            args_types=list(map(decode_type, data['args_types']))
        )

    def set_implementation(self, new_imp: int):
        self.client.symbols.method_setImplementation(self.address, new_imp)
        self.imp = self.client.symbol(new_imp)

    def always_return(self, value: bool) -> None:
        """
        Patch the method to always return the given value.
        """
        if value:
            # Exported from rpcserver binary
            ret_val = self.client.symbols.get_true
        else:
            # Exported from rpcserver binary
            ret_val = self.client.symbols.get_false
        self.set_implementation(ret_val)

    def __str__(self):
        if ':' in self.name:
            args_names = self.name.split(':')
            name = ' '.join(['{}:({})'.format(*arg) for arg in zip(args_names, self.args_types[2:])])
        else:
            name = self.name
        prefix = '+' if self.is_class else '-'
        return f'{prefix} {name}; // 0x{self.address:x} (returns: {self.return_type})\n'



import datetime

import pytest

pytestmark = pytest.mark.darwin


@pytest.mark.parametrize('data', [
    None,
    True,
    False,
    datetime.datetime(1967, 5, 4, 3, 2, 1),
    'string',
    b'bytes',
    123,
    0.1,
    [0, 1, 'abc'],
    {'key': 'value'},
    [{'key': 'value'}, [1, 2]]])
def test_serialization(client, data):
    assert client.cf(data).py() == data



O_RDONLY = 0
O_WRONLY = 1
O_RDWR = 2
O_CREAT = 512
O_TRUNC = 1024

R_OK = 0x04

SEEK_SET = 0
SEEK_CUR = 1
SEEK_END = 2
SEEK_HOLE = 3
SEEK_DATA = 4

S_IFMT = 0o00170000
S_IFSOCK = 0o0140000
S_IFLNK = 0o0120000
S_IFREG = 0o0100000
S_IFBLK = 0o0060000
S_IFDIR = 0o0040000
S_IFCHR = 0o0020000
S_IFIFO = 0o0010000
S_ISUID = 0o0004000
S_ISGID = 0o0002000
S_ISVTX = 0o0001000

SOL_IP = 0
SOL_SOCKET = 65535
SOL_TCP = 6
SOL_UDP = 17

SO_SNDLOWAT = 0x1003
SO_RCVLOWAT = 0x1004
SO_SNDTIMEO = 0x1005
SO_RCVTIMEO = 0x1006

MSG_WAITALL = 0x40
MSG_NOSIGNAL = 524288

AF_UNIX = 1
AF_INET = 2
AF_INET6 = 30

SOCK_STREAM = 1

IPPROTO_TCP = 6

DT_UNKNOWN = 0
DT_FIFO = 1
DT_CHR = 2
DT_DIR = 4
DT_BLK = 6
DT_REG = 8
DT_LNK = 10
DT_SOCK = 12
DT_WHT = 14

RTLD_GLOBAL = -2
RTLD_NEXT = -1
RTLD_NOW = 2

SIGABRT = 6
SIGALRM = 14
SIGBUS = 10
SIGCHLD = 20
SIGCONT = 19
SIGEMT = 7
SIGFPE = 8
SIGHUP = 1
SIGILL = 4
SIGINFO = 29
SIGINT = 2
SIGIO = 23
SIGIOT = 6
SIGKILL = 9
SIGPIPE = 13
SIGPROF = 27
SIGQUIT = 3
SIGSEGV = 11
SIGSTOP = 17
SIGSYS = 12
SIGTERM = 15
SIGTRAP = 5
SIGTSTP = 18
SIGTTIN = 21
SIGTTOU = 22
SIGURG = 16
SIGUSR1 = 30
SIGUSR2 = 31
SIGVTALRM = 26
SIGWINCH = 28
SIGXCPU = 24
SIGXFSZ = 25

EPERM = 1
ENOENT = 2
EACCESS = 13
EEXIST = 17
ENOTDIR = 20
EISDIR = 21
EPIPE = 32
EAGAIN = 35
ECONNREFUSED = 61
ENOTEMPTY = 66

F_SETFL = 4
F_GETFL = 3

O_NONBLOCK = 4



import ctypes
import os
import struct
from contextlib import contextmanager
from typing import List

from capstone import CS_ARCH_ARM64, CS_ARCH_X86, CS_MODE_64, CS_MODE_LITTLE_ENDIAN, Cs, CsInsn
from construct import Container

from rpcclient.protos.rpc_pb2 import ARCH_ARM64
from rpcclient.structs.generic import Dl_info

ADDRESS_SIZE_TO_STRUCT_FORMAT = {1: 'B', 2: 'H', 4: 'I', 8: 'Q'}
RETVAL_BIT_COUNT = 64


class Symbol(int):
    """ wrapper for a remote symbol object """

    PROXY_METHODS = ['peek', 'poke']

    @classmethod
    def create(cls, value: int, client):
        """
        Create a Symbol object.
        :param value: Symbol address.
        :param rpcclient.darwin_client.Client client: client.
        :return: Symbol object.
        :rtype: Symbol
        """
        if not isinstance(value, int):
            raise TypeError()

        value &= 0xFFFFFFFFFFFFFFFF

        symbol = cls(value)
        symbol._prepare(client)
        return symbol

    def _clone_from_value(self, value: int):
        return self.create(value, self._client)

    def _prepare(self, client):
        self.retval_bit_count = RETVAL_BIT_COUNT
        self.is_retval_signed = True
        self.item_size = 8

        # private members
        self._client = client
        self._offset = 0

        for method_name in Symbol.PROXY_METHODS:
            getattr(self.__class__, method_name).__doc__ = \
                getattr(client, method_name).__doc__

    @contextmanager
    def change_item_size(self, new_item_size: int):
        """
        Temporarily change item size
        :param new_item_size: Temporary item size
        """
        save_item_size = self.item_size
        self.item_size = new_item_size
        try:
            yield
        finally:
            self.item_size = save_item_size

    def peek(self, count):
        return self._client.peek(self, count)

    def poke(self, buf):
        return self._client.poke(self, buf)

    def peek_str(self, encoding='utf-8') -> str:
        """ peek string at given address """
        return self.peek(self._client.symbols.strlen(self)).decode(encoding)

    def close(self):
        """ Construct compliance. """
        pass

    def seek(self, offset, whence):
        """ Construct compliance. """
        if whence == os.SEEK_CUR:
            self._offset += offset
        elif whence == os.SEEK_SET:
            self._offset = offset - self
        else:
            raise IOError('Unsupported whence')

    def read(self, count: int):
        """ Construct compliance. """
        val = (self + self._offset).peek(count)
        self._offset += count
        return val

    def write(self, buf):
        """ Construct compliance. """
        val = (self + self._offset).poke(buf)
        self._offset += len(buf)
        return val

    def tell(self):
        """ Construct compliance. """
        return self + self._offset

    def disass(self, size=40) -> List[CsInsn]:
        """ peek disassembled lines of 'size' bytes """
        if self._client.arch == ARCH_ARM64:
            return list(Cs(CS_ARCH_ARM64, CS_MODE_LITTLE_ENDIAN).disasm(self.peek(size), self))
        else:
            # assume x86_64 by default
            return list(Cs(CS_ARCH_X86, CS_MODE_LITTLE_ENDIAN | CS_MODE_64).disasm(self.peek(size), self))

    @property
    def c_int64(self) -> int:
        """ cast to c_int64 """
        return ctypes.c_int64(self).value

    @property
    def c_uint64(self) -> int:
        """ cast to c_uint64 """
        return ctypes.c_uint64(self).value

    @property
    def c_int32(self) -> int:
        """ cast to c_int32 """
        return ctypes.c_int32(self).value

    @property
    def c_uint32(self) -> int:
        """ cast to c_uint32 """
        return ctypes.c_uint32(self).value

    @property
    def c_int16(self) -> int:
        """ cast to c_int16 """
        return ctypes.c_int16(self).value

    @property
    def c_uint16(self) -> int:
        """ cast to c_uint16 """
        return ctypes.c_uint16(self).value

    @property
    def c_bool(self) -> bool:
        """ cast to c_bool """
        return ctypes.c_bool(self).value

    @property
    def dl_info(self) -> Container:
        dl_info = Dl_info(self._client)
        sizeof = dl_info.sizeof()
        with self._client.safe_malloc(sizeof) as info:
            if 0 == self._client.symbols.dladdr(self, info):
                self._client.raise_errno_exception(f'failed to extract info for: {self}')
            return dl_info.parse_stream(info)

    @property
    def name(self) -> str:
        return self.dl_info.dli_sname

    @property
    def filename(self) -> str:
        return self.dl_info.dli_fname

    def __add__(self, other):
        try:
            return self._clone_from_value(int(self) + other)
        except TypeError:
            return int(self) + other

    def __radd__(self, other):
        return self.__add__(other)

    def __sub__(self, other):
        try:
            return self._clone_from_value(int(self) - other)
        except TypeError:
            return int(self) - other

    def __rsub__(self, other):
        try:
            return self._clone_from_value(other - int(self))
        except TypeError:
            return other - int(self)

    def __mul__(self, other):
        try:
            return self._clone_from_value(int(self) * other)
        except TypeError:
            return int(self) * other

    def __rmul__(self, other):
        return self.__mul__(other)

    def __truediv__(self, other):
        return self._clone_from_value(int(self) / other)

    def __floordiv__(self, other):
        return self._clone_from_value(int(self) // other)

    def __mod__(self, other):
        return self._clone_from_value(int(self) % other)

    def __and__(self, other):
        return self._clone_from_value(int(self) & other)

    def __or__(self, other):
        return self._clone_from_value(int(self) | other)

    def __xor__(self, other):
        return self._clone_from_value(int(self) ^ other)

    def __getitem__(self, item):
        fmt = ADDRESS_SIZE_TO_STRUCT_FORMAT[self.item_size]
        addr = self + item * self.item_size
        return self._clone_from_value(
            struct.unpack(self._client._endianness + fmt, self._client.peek(addr, self.item_size))[0])

    def __setitem__(self, item, value):
        fmt = ADDRESS_SIZE_TO_STRUCT_FORMAT[self.item_size]
        value = struct.pack(self._client._endianness + fmt, int(value))
        self._client.poke(self + item * self.item_size, value)

    def __repr__(self):
        return f'<{type(self).__name__}: {hex(self)}>'

    def __str__(self):
        return hex(self)

    def __call__(self, *args, **kwargs):
        return self._client.call(self, args, **kwargs)



import datetime
import logging
import re
from typing import List

from cached_property import cached_property

from rpcclient.darwin.cfpreferences import GLOBAL_DOMAIN, kCFPreferencesAnyHost
from rpcclient.darwin.consts import OsLogLevel
from rpcclient.darwin.processes import Process
from rpcclient.darwin.symbol import DarwinSymbol
from rpcclient.exceptions import BadReturnValueError, HarGlobalNotFoundError, MissingLibraryError
from rpcclient.structs.consts import SIGKILL

MOV_RDI_RBX = b'\x48\x89\xdf'
XOR_ESI_ESI = b'\x31\xf6'
MOV_EDX_0A = b'\xba\x0a\x00\x00'
INTEL_PATTERN = MOV_RDI_RBX + XOR_ESI_ESI + MOV_EDX_0A + b'\x00'

MOV_X0_X9 = b'\xE0\x03\x13\xAA'
MOV_X1_0 = b'\x01\x00\x80\xD2'
MOV_W2_0XA = b'\x42\x01\x80\x52'
ARM_PATTERN = MOV_X0_X9 + MOV_X1_0 + MOV_W2_0XA

logger = logging.getLogger(__name__)


class OsLogPreferencesBase:
    def __init__(self, client, obj: DarwinSymbol):
        self._client = client
        self._object = obj

    @property
    def name(self) -> str:
        return self._object.objc_call('name').py()

    @property
    def persisted_level(self) -> OsLogLevel:
        return OsLogLevel(self._object.objc_call('persistedLevel'))

    @persisted_level.setter
    def persisted_level(self, value: OsLogLevel) -> None:
        self._object.objc_call('setPersistedLevel:', value)

    @property
    def default_persisted_level(self) -> OsLogLevel:
        return OsLogLevel(self._object.objc_call('defaultPersistedLevel'))

    @property
    def enabled_level(self) -> OsLogLevel:
        return OsLogLevel(self._object.objc_call('enabledLevel'))

    @enabled_level.setter
    def enabled_level(self, value: OsLogLevel) -> None:
        self._object.objc_call('setEnabledLevel:', value)

    @property
    def default_enabled_level(self) -> OsLogLevel:
        return OsLogLevel(self._object.objc_call('defaultEnabledLevel'))

    @property
    def _verbosity_description(self) -> str:
        return (f'ENABLED_LEVEL:{self.enabled_level.name}({self.enabled_level.value}) '
                f'PERSISTED_LEVEL:{self.persisted_level.name}({self.persisted_level.value})')

    def reset(self) -> None:
        self._object.objc_call('reset')

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__} NAME:{self.name} {self._verbosity_description}>'


class OsLogPreferencesCategory(OsLogPreferencesBase):
    def __init__(self, client, category: str, subsystem: DarwinSymbol):
        obj = client.symbols.objc_getClass('OSLogPreferencesCategory').objc_call('alloc').objc_call(
            'initWithName:subsystem:', client.cf(category), subsystem)
        super().__init__(client, obj)

    @property
    def name(self) -> str:
        return self._object.objc_call('name').py()

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__} NAME:{self.name} {self._verbosity_description}>'


class OsLogPreferencesSubsystem(OsLogPreferencesBase):
    def __init__(self, client, subsystem: str):
        obj = client.symbols.objc_getClass('OSLogPreferencesSubsystem').objc_call('alloc').objc_call(
            'initWithName:', client.cf(subsystem))
        super().__init__(client, obj)

    @property
    def category_strings(self) -> List[OsLogPreferencesCategory]:
        return self._object.objc_call('categories').py()

    @property
    def categories(self) -> List[OsLogPreferencesCategory]:
        result = []
        for name in self.category_strings:
            result.append(self.get_category(name))
        return result

    def get_category(self, name: str) -> OsLogPreferencesCategory:
        return OsLogPreferencesCategory(self._client, name, self._object)


class OsLogPreferencesManager(OsLogPreferencesBase):
    def __init__(self, client):
        obj = client.symbols.objc_getClass('OSLogPreferencesManager').objc_call('sharedManager')
        super().__init__(client, obj)

    @property
    def subsystem_strings(self) -> List[str]:
        return self._object.objc_call('subsystems').py()

    @property
    def subsystems(self) -> List[OsLogPreferencesSubsystem]:
        result = []
        for name in self.subsystem_strings:
            result.append(self.get_subsystem(name))
        return result

    def get_subsystem(self, name: str) -> OsLogPreferencesSubsystem:
        return OsLogPreferencesSubsystem(self._client, name)


class Syslog:
    """" manage syslog """

    def __init__(self, client):
        """
        @type client: rpcclient.darwin.client.DarwinClient
        """
        self._client = client
        self._load_logging_support_library()
        self.preferences_manager = OsLogPreferencesManager(self._client)

    @cached_property
    def _cfnetwork_base(self) -> int:
        self._client.load_framework('CFNetwork')
        cfnetwork = [image for image in self._client.images if
                     image.name.startswith('/System/Library/Frameworks/CFNetwork.framework')]
        if len(cfnetwork) < 1:
            raise MissingLibraryError()
        return cfnetwork[0].base_address

    @cached_property
    def _arm_enable_har_global(self) -> DarwinSymbol:
        """
        In order to find an unexported global variable, we use a sequence of instructions that we know comes after
        the aforementioned variable is accessed.

            88 07 2D D0                 ADRP            X8, #har_global@PAGE
            08 41 55 39                 LDRB            W8, [X8,#har_global@PAGEOFF]
            08 03 00 34                 CBZ             W8, loc_187DA1294
            E0 03 13 AA                 MOV             X0, X19 ; __str
            01 00 80 D2                 MOV             X1, #0  ; __endptr
            42 01 80 52                 MOV             W2, #0xA ; __base

        After getting the address of the instruction `mov x0,x19`, we need to rewind 2 instructions. Next, we verify
        that the above 2 instructions are `adrp` && `ldrb` . Then, extract the page address + page offset and
        calculate the affective address.
        """

        address = []
        regex_hex = r'0x[0-9a-fA-F]+'
        start = self._cfnetwork_base
        while True:
            pattern_addr = self._client.symbols.memmem(start, 0xffffffff, ARM_PATTERN, len(ARM_PATTERN))
            if pattern_addr == 0:
                raise HarGlobalNotFoundError()
            pattern_sym = self._client.symbol(pattern_addr)
            disass = (pattern_sym - len(ARM_PATTERN)).disass(8)
            if disass[0].mnemonic == 'adrp' and disass[1].mnemonic == 'ldrb':
                break
            # search next occurrence
            start = pattern_addr + 1

        for instruction in disass:
            address.append(
                int(re.findall(regex_hex, instruction.op_str)[0], 16)
            )
        return self._client.symbol(sum(address))

    @cached_property
    def _intel_enable_har_global(self) -> DarwinSymbol:
        """
        In order to find an unexported global variable, we use a sequence of instructions that we know comes before
        the aforementioned variable is accessed.

            48 89 DF                    mov     rdi, rbx        ; __str
            31 F6                       xor     esi, esi        ; __endptr
            BA 0A 00 00                 mov     edx, 0Ah        ; __base
            00
            E8 78 EE 10                 call    _strtol
            00
            48 89 C3                    mov     rbx, rax
            88 1D 19 9C                 mov     cs:har_global_0, bl  <-----------------
        """
        op_str_pattern = 'byte ptr [rip +'
        start = self._cfnetwork_base
        while True:
            pattern_addr = self._client.symbols.memmem(start, 0xffffffff, INTEL_PATTERN, len(INTEL_PATTERN))
            if pattern_addr == 0:
                raise HarGlobalNotFoundError()
            pattern_sym = self._client.symbol(pattern_addr)
            disass = (pattern_sym + len(INTEL_PATTERN)).disass(100)
            if (disass[0].mnemonic == 'call' and disass[1].mnemonic == 'mov' and disass[2].mnemonic == 'mov' and
                    disass[2].op_str.startswith(op_str_pattern)):
                offset = int(disass[2].op_str.split(op_str_pattern, 1)[1].split(']')[0], 0)
                return self._client.symbol(disass[2].address + offset)
            start = pattern_addr + 1

    @cached_property
    def _enable_har_global(self) -> DarwinSymbol:
        if self._client.uname.machine == 'x86_64':
            return self._intel_enable_har_global
        return self._arm_enable_har_global

    def install_cfnetwork_diagnostics_profile(self) -> None:
        self._client.preferences.cf.set('AppleCFNetworkDiagnosticLogging', 3, GLOBAL_DOMAIN,
                                        username='kCFPreferencesAnyUser')

    def remove_cfnetwork_diagnostics_profile(self) -> None:
        self._client.preferences.cf.remove('AppleCFNetworkDiagnosticLogging', GLOBAL_DOMAIN,
                                           username='kCFPreferencesAnyUser')

    def set_harlogger_for_process(self, value: bool, pid: int) -> None:
        process = self._client.processes.get_by_pid(pid)
        self._set_harlogger_for_process(value, process)
        self.set_har_capture_global(True)

    def set_harlogger_for_all(self, value: bool, expression: str = None) -> None:
        for p in self._client.processes.list():
            if p.pid == 0:
                continue
            if expression and expression not in p.basename:
                continue
            try:
                self._set_harlogger_for_process(value, p)
                logger.info(f'{"Enabled" if value else "Disabled"} for {p.basename}')
            except BadReturnValueError:
                logger.error(f'Failed To enable for {p}')
        self.set_har_capture_global(True)

    def set_unredacted_logs(self, enable: bool = True):
        """
        enable/disable unredacted logs (allows seeing the <private> strings)
        https://github.com/EthanArbuckle/unredact-private-os_logs
        """
        with self._client.preferences.sc.open(
                '/Library/Preferences/Logging/com.apple.system.logging.plist') as pref:
            pref.set_dict({'Enable-Logging': True, 'Enable-Private-Data': enable})
        self._client.processes.get_by_basename('logd').kill(SIGKILL)

    def set_har_capture_global(self, enable: bool = True):
        """
        enable/disable HAR logging
        https://github.com/doronz88/harlogger
        """
        users = ['mobile', 'root']
        if enable:
            for user in users:
                # settings copied from DiagnosticExtension.appex
                self._client.preferences.cf.set('har-capture-global',
                                                datetime.datetime(9999, 12, 31, 23, 59, 59),
                                                'com.apple.CFNetwork', user, hostname=kCFPreferencesAnyHost)
                self._client.preferences.cf.set('har-body-size-limit',
                                                0x100000,
                                                'com.apple.CFNetwork', user, hostname=kCFPreferencesAnyHost)
                subsystem = self.preferences_manager.get_subsystem('com.apple.CFNetwork')
                category = subsystem.get_category('HAR')

                category.enabled_level = 2
                category.persisted_level = 2
        else:
            for user in users:
                self._client.preferences.cf.set('har-capture-global',
                                                datetime.datetime(1970, 1, 1, 1, 1, 1),
                                                'com.apple.CFNetwork', user, hostname=kCFPreferencesAnyHost)

        if self._client.symbols.notify_post('com.apple.CFNetwork.har-capture-update'):
            raise BadReturnValueError('notify_post() failed')

    def _load_logging_support_library(self) -> None:
        self._client.load_framework('LoggingSupport')

    def _set_harlogger_for_process(self, value: bool, process: Process) -> None:
        process.poke(self._enable_har_global, value.to_bytes(1, 'little'))
